# -*- coding: utf-8 -*-

from b3j0f.conf import Configurable, category, Parameter

from link.middleware.core import register_middleware
from link.kvstore.driver import Driver

from link.riak.features.model.core import RiakSolrSchema
from link.riak.features.fulltext import RiakSearch2
from link.feature import addfeatures
from link.riak import CONF_BASE_PATH

from six import string_types, raise_from
import riak


@Configurable(
    paths='{0}/driver.conf'.format(CONF_BASE_PATH),
    conf=category(
        'RIAK',
        Parameter(name='default_bucket', value='default'),
        Parameter(name='indexing', ptype=bool, value=False),
        Parameter(name='protocol', value='http'),
        Parameter(name='pkey'),
        Parameter(name='cert'),
        Parameter(name='cacert'),
        Parameter(name='crl'),
        Parameter(name='ciphers'),
        Parameter(name='sslver', ptype=int)
    )
)
@addfeatures([RiakSolrSchema, RiakSearch2])
@register_middleware
class RiakDriver(Driver):

    __protocols__ = ['riak']

    def __init__(
        self,
        default_bucket=None,
        indexing=False,
        protocol=None,
        pkey=None,
        cert=None,
        cacert=None,
        crl=None,
        ciphers=None,
        sslver=5,
        *args, **kwargs
    ):
        super(RiakDriver, self).__init__(*args, **kwargs)

        self.default_bucket = default_bucket
        self.indexing = indexing
        self.protocol = protocol
        self.pkey = pkey
        self.cert = cert
        self.cacert = cacert
        self.crl = crl
        self.ciphers = ciphers
        self.sslver = sslver

    def _connect(self):
        nodes = []
        security = None

        for host, port in self.hosts:
            node = {
                'host': host
            }

            if self.protocol == 'pbc':
                node['pb_port'] = port

            else:
                node['http_port'] = port

            nodes.append(node)

        kwargs = {}

        if self.user is not None:
            kwargs['username'] = self.user

            if self.pwd:
                kwargs['password'] = self.pwd

        if self.pkey is not None:
            kwargs['pkey_file'] = self.pkey

        if self.cert is not None:
            kwargs['cert_file'] = self.cert

        if self.cacert is not None:
            kwargs['cacert_file'] = self.cacert

        if self.crl is not None:
            kwargs['crl_file'] = self.crl

        if self.ciphers is not None:
            kwargs['ciphers'] = self.ciphers

        if kwargs:
            kwargs['ssl_version'] = self.sslver

            security = riak.security.SecurityCreds(**kwargs)

        else:
            security = None

        return riak.RiakClient(nodes=nodes, credentials=security)

    def _disconnect(self, conn):
        conn.close()

    def _isconnected(self, conn):
        return conn is not None and conn.is_alive()

    def _get_bucket(self, conn):
        if len(self.path) == 0:
            bucket = conn.bucket(self.default_bucket)

        elif len(self.path) == 1:
            bucket = conn.bucket(self.path[0])

        else:
            bucket_type = conn.bucket_type(self.path[0])
            bucket = bucket_type.bucket('_'.join(self.path[1:]))

        return bucket

    def _get_indexes(self, val, prefix=''):
        pairs = []

        if isinstance(val, int):
            pairs.append(('{0}_int'.format(prefix), val))

        elif isinstance(val, string_types):
            pairs.append(('{0}_bin'.format(prefix), val))

        elif isinstance(val, dict):
            for key in val:
                pairs += self._get_indexes(
                    val[key],
                    '{0}_{1}'.format(prefix, key)
                )

        elif isinstance(val, list):
            for i in range(len(val)):
                item = val[i]

                pairs += self._get_indexes(
                    item,
                    '{0}_{1}'.format(prefix, i)
                )

        return pairs

    def _new_object(self, conn, key, val):
        obj = self._get_bucket(conn).new(key, val)

        if self.indexing:
            for indexkey, indexval in self._get_indexes(val):
                obj.add_index(indexkey, indexval)

        return obj

    def _get(self, conn, key):
        obj = self._get_bucket(conn).get(key)

        if not obj.exists:
            raise KeyError('No such key: {0}'.format(key))

        return obj.data

    def _multiget(self, conn, keys):
        bucket = self._get_bucket(conn)

        results = bucket.multiget(keys)

        datas = []

        for result in results:
            if isinstance(result, riak.RiakObject):
                datas.append(result.data)

            elif isinstance(result, riak.datatypes.Datatype):
                doc = {}

                for subkey, type_subkey in result.value:
                    dockey = '{0}_{1}'.format(subkey, type_subkey)
                    val = result.value[(subkey, type_subkey)]

                    if isinstance(val, frozenset):
                        val = list(val)

                    else:
                        val = str(val)

                    doc[dockey] = val

                datas.append(doc)

            else:
                _, _, key, err = result

                raise_from(
                    KeyError('No such key: {0}'.format(key)),
                    err
                )

        return datas

    def _put(self, conn, key, val):
        obj = self._new_object(conn, key, val)
        obj.store()

    def _multiput(self, conn, keys, vals):
        objs = [
            self._new_object(conn, k, v)
            for k, v in zip(keys, vals)
        ]
        conn.multiput(objs)

    def _remove(self, conn, key):
        obj = self._get_bucket(conn).get(key)

        if not obj.exists:
            raise KeyError('No such key: {0}'.format(key))

        if self.indexing:
            for indexkey, indexval in self._get_indexes(obj.data):
                obj.remove_index(indexkey, indexval)

        obj.delete()

    def _exists(self, conn, key):
        obj = self._get_bucket(conn).get(key)
        return obj.exists

    def _keys(self, conn):
        return self._get_bucket(conn).get_keys()



# -*- coding: utf-8 -*-

from b3j0f.conf import Configurable

from xml.etree import ElementTree as ET
from copy import deepcopy

from link.riak.features.model.utils import create_model_class
from link.riak.features.model.utils import get_member
from link.riak.features.model.utils import get_type
from link.model.feature import ModelFeature, Model
from link.riak import CONF_BASE_PATH


@Configurable(paths='{0}/model.conf'.format(CONF_BASE_PATH))
class RiakSolrSchema(ModelFeature):

    DATA_ID = '_yz_rk'

    def create_model(self, schema):
        root = ET.fromstring(schema)

        clsname = root.get('name')
        clsbases = (Model,)
        clsmembers = {}

        typemapping = {}

        for _type in root.iter('fieldType'):
            typename = _type.get('name')
            typecls = get_type(_type.get('class'))
            typeattr = deepcopy(_type.attrib)

            typeattr.pop('name')
            typeattr.pop('class')

            typemapping[typename] = (typecls, typeattr)

        for field in root.iter('field'):
            fname = field.get('name')
            fnames = fname.split('.')

            if len(fnames) == 1:
                clsmembers[fname] = get_member(typemapping, field)

            else:
                tree = clsmembers

                for name in fnames[:-1]:
                    if name not in tree:
                        tree[name] = {}

                    tree = tree[name]

                tree[fnames[-1]] = get_member(typemapping, field)

        return create_model_class(clsname, clsbases, clsmembers)



# -*- coding: utf-8 -*-

from riak.datatypes import Register, Map, Counter, Flag, Set
from b3j0f.utils.iterable import isiterable
from b3j0f.aop import weave, get_advices
from six import string_types

from link.riak.features.model.utils import create_model_class
from link.model.feature import Model


class BaseField(property):
    @property
    def name(self):
        return self._attr['name']

    @property
    def attr(self):
        return '_{0}'.format(self._attr['name'])

    @property
    def isregister(self):
        return self.name.endswith('_register')

    @property
    def ismap(self):
        return self.name.endswith('_map')

    @property
    def iscounter(self):
        return self.name.endswith('_counter')

    @property
    def isflag(self):
        return self.name.endswith('_flag')

    @property
    def isset(self):
        return self.name.endswith('_set')

    @property
    def is_datatype(self):
        return any([
            self.isregister,
            self.ismap,
            self.iscounter,
            self.isflag,
            self.isset
        ])

    @property
    def dtkey(self):
        result = self.name

        if self.isregister:
            result = result[:-len('_register')]

        elif self.ismap:
            result = result[:-len('_map')]

        elif self.iscounter:
            result = result[:-len('_counter')]

        elif self.isflag:
            result = result[:-len('_flag')]

        elif self.isset:
            result = result[:-len('_set')]

        return result

    def __init__(self, attributes, *args, **kwargs):
        kwargs['fget'] = self._fget
        kwargs['fset'] = self._fset
        kwargs['fdel'] = self._fdel

        super(BaseField, self).__init__(*args, **kwargs)

        self._attr = attributes

    def convert_value(self, val):
        raise NotImplementedError()

    def setdefault(self, obj):
        raise NotImplementedError()

    def _fget(self, obj):
        try:
            result = getattr(obj, self.attr)

        except AttributeError:
            result = self.setdefault(obj)

        if self.isregister:
            if not isinstance(result, Register):
                obj._map.registers[self.dtkey].assign(result)
                result = obj._map.registers[self.dtkey]

        elif self.ismap:
            pass

        elif self.iscounter:
            if not isinstance(result, Counter):
                if result >= 0:
                    obj._map.counters[self.dtkey].increment(result)

                else:
                    obj._map.counters[self.dtkey].decrement(-result)

                result = obj._map.counters[self.dtkey]

        elif self.isflag:
            if not isinstance(result, Flag):
                if result:
                    obj._map.flags[self.dtkey].enable()

                else:
                    obj._map.flags[self.dtkey].disable()

                result = obj._map.flags[self.dtkey]

        elif self.isset:
            if not isinstance(result, Set):
                for item in result:
                    obj._map.sets[self.dtkey].add(item)

                result = obj._map.sets[self.dtkey]

        return result

    def _fset(self, obj, val):
        val = self.convert_value(val)

        if self.is_datatype:
            dt = self._fget(obj)

            if self.isregister:
                dt.assign(val)

            elif self.iscounter:
                v = val - (dt.value + dt._increment)

                if v >= 0:
                    dt.increment(v)

                else:
                    dt.decrement(-v)

            elif self.isflag:
                if v:
                    dt.enable()

                else:
                    dt.disable()

            elif self.isset:
                for item in val:
                    if item not in dt.value:
                        dt.add(item)

                    else:
                        dt.discard(item)

            elif self.ismap:
                raise AttributeError('Cannot set map attributes')

            val = dt

        setattr(obj, self.attr, val)

    def _fdel(self, obj):
        if self.is_datatype:
            if self.ismap:
                obj._map.maps[self.dtkey].delete()

            else:
                dt = self._fget(obj)
                dt.delete()

        else:
            delattr(obj, self.attr)


class EmbeddedModelField(BaseField):
    def __init__(self, members, *args, **kwargs):
        super(EmbeddedModelField, self).__init__(*args, **kwargs)

        self.cls = create_model_class(
            'EmbeddedModel_{0}'.format(self.name),
            (Model,),
            members
        )

    def convert_value(self, val):
        if not isinstance(val, self.cls):
            raise TypeError('{0} must be an embedded model: {1}'.format(
                self.name,
                self.cls.__name__
            ))

        return val

    def setdefault(self, obj):
        result = self.cls(
            obj._middleware, '',
            model_map=obj._map.maps[self.dtkey]
        )
        setattr(obj, self.attr, result)
        return result


class TrieIntField(BaseField):
    def convert_value(self, val):
        if not isinstance(val, int):
            val = int(val)

        return val

    def setdefault(self, obj):
        result = 0
        setattr(obj, self.attr, result)
        return result


class BoolField(BaseField):
    def convert_value(self, val):
        if not isinstance(val, bool):
            val = (val == 'true')

        return val

    def setdefault(self, obj):
        result = False
        setattr(obj, self.attr, result)
        return result


class StrField(BaseField):
    def convert_value(self, val):
        if not isinstance(val, string_types):
            val = str(val)

        return val

    def setdefault(self, obj):
        result = ''
        setattr(obj, self.attr, result)
        return result


class ArrayField(BaseField):
    def __init__(self, subtype, *args, **kwargs):
        super(ArrayField, self).__init__(*args, **kwargs)

        self.subtype = subtype(self.attr)

    def ensure_type_append(self, jointpoint):
        args = list(jointpoint.args)
        args[0] = self.subtype.convert_value(args[0])
        jointpoint.args = tuple(args)

        return jointpoint.proceed()

    def ensure_type_extend(self, jointpoint):
        L = jointpoint.args[0]

        for i, item in enumerate(L):
            L[i] = self.subtype.convert_value(item)

        return jointpoint.proceed()

    def ensure_type_insert(self, jointpoint):
        args = list(jointpoint)
        args[1] = self.subtype.convert_value(args[1])
        jointpoint.args = tuple(args)

        return jointpoint.proceed()

    def weave_list(self, val):
        methods = [
            (val.append, self.ensure_type_append),
            (val.extend, self.ensure_type_extend),
            (val.insert, self.ensure_type_insert),
            (val.remove, self.ensure_type_append),
            (val.__setitem__, self.ensure_type_insert)
        ]

        for method, advice in methods:
            if advice not in get_advices(method):
                weave(target=method, advices=advice)

    def setdefault(self, obj):
        result = []
        self.weave_list(result)
        setattr(obj, self.attr, result)
        return result

    def convert_value(self, val):
        if not isiterable(val, exclude=string_types):
            val = [val]

        self.weave_list(val)

        return val



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
import os
import re


NAME = 'link.riak'
KEYWORDS = 'link riak'
DESC = 'Riak key/value store driver'
URL = 'https://github.com/linkdd/link.riak'
AUTHOR = 'David Delassus'
AUTHOR_EMAIL = 'david.jose.delassus@gmail.com'
LICENSE = 'MIT'
REQUIREMENTS = [
    'b3j0f.conf>=0.3.18',
    'b3j0f.aop>=0.8.4',
    'b3j0f.utils>=1.4.3',
    'link.feature>=2.0',
    'link.middleware>=1.4',
    'link.kvstore>=0.10',
    'link.fulltext>=0.6',
    'link.model>=0.8',
    'six>=1.10.0',
    'riak>=2.5.5'
]

CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Topic :: Utilities',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Intended Audience :: Developers',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: Implementation :: CPython'
]


def get_cwd():
    return os.path.dirname(os.path.abspath(os.path.expanduser(__file__)))


def get_version(default='0.1'):
    _name = NAME.replace('.', os.sep)
    path = os.path.join(get_cwd(), _name, '__init__.py')

    with open(path) as f:
        stream = f.read()
        regex = re.compile(r'.*__version__ = \'(.*?)\'', re.S)
        version = regex.match(stream)

        if version is None:
            version = default

        else:
            version = version.group(1)

    return version


def get_long_description():
    path = os.path.join(get_cwd(), 'README.rst')
    desc = None

    if os.path.exists(path):
        with open(path) as f:
            desc = f.read()

    return desc


def get_scripts():
    path = os.path.join(get_cwd(), 'scripts')
    scripts = []

    if os.path.exists(path):
        for root, _, files in os.walk(path):
            for f in files:
                scripts.append(os.path.join(root, f))

    return scripts


def get_test_suite():
    return 'link'


setup(
    name=NAME,
    keywords=KEYWORDS,
    version=get_version(),
    url=URL,
    description=DESC,
    long_description=get_long_description(),
    license=LICENSE,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    packages=find_packages(),
    scripts=get_scripts(),
    test_suite=get_test_suite(),
    install_requires=REQUIREMENTS,
    classifiers=CLASSIFIERS
)



# -*- coding: utf-8 -*-

from b3j0f.utils.path import lookup
from riak.datatypes import Map
from copy import deepcopy


def get_type(name):
    return lookup('link.riak.features.model.{0}'.format(name))


def get_member(typemapping, field):
    ftype, typeattr = typemapping.get(field.get('type'))
    fattr = deepcopy(field.attrib)
    fattr.update(typeattr)

    isarray = (field.get('multiValued', 'false') == 'true')

    if isarray:
        ArrayField = get_type('solr.ArrayField')
        member = ArrayField(ftype, fattr)

    else:
        member = ftype(fattr)

    return member


def model_save(model):
    model._map.key = model[model._DATA_ID] if model._DATA_ID in model else None

    try:
        model._middleware.conn.update_datatype(model._map)

    except ValueError:  # No operation to send
        pass

    else:
        model[model._DATA_ID] = model._map.key


def model_delete(model):
    model._middleware.conn.delete(model._map)

    if model._DATA_ID in model:
        del model[model._DATA_ID]


def create_model_class(name, bases, members):
    def model_init(model, model_map=None, *args, **kwargs):
        for base in bases:
            base.__init__(model, *args, **kwargs)

        if model_map is None:
            model_map = Map(
                bucket=model._middleware.conn.bucket_type(
                    '{0}s'.format(model._schemaname)
                ).bucket(
                    'default'
                )
            )

        model._map = model_map

    clsmembers = {
        '_DATA_ID': '_yz_rk',
        '__init__': model_init,
        'save': model_save,
        'delete': model_delete
    }

    for mname in members:
        member = members[mname]

        if isinstance(member, dict):
            EmbeddedModelField = get_type('solr.EmbeddedModelField')

            member = EmbeddedModelField(
                member,
                {
                    'name': mname
                }
            )

        clsmembers[mname] = member

    return type(name, bases, clsmembers)
